// Ytruder.cpp : Defines the entry point for the console application.
//

// LDraw line extruder 
// Extrudes a surface from a path (ordered sequence of LDraw type 2 lines).
// Extrusion occurs along specified LDraw axis.
// Optional lines are set up between triangles and quads. End of paths receive edge lines.
// If angle between generated surfaces exceed some threshold, conditional lines are converted to edge lines (type 2)

// Usage: ytruder [-d or -s or -p <Yval>] [-x or -y or -z]
//                [-a <angle>] LdrawLineFile LdrawSurfaceFileOut
// Options
//   -x : Operates in x direction
//   -y : Operates in y direction (default)
//   -z : Operates in z direction
//   -d <val>: extrude lines by a distance <val>
//   -s <val>: extrude lines symmetrically vs. plane <val>.
//   -p <val>: extrude down to plane <val> (projection).
//   -r <val>: extrude radially by a distance <val> from direction axis.
//              Default for -d/s/p <val> is -s 0 (symmetry vs. origin plane).
//   -a: Conditional to fixed line conversion angle threshold <angle>.
//       Default value is 30 degrees.

// If input files are not ordered they will be ordered first.
// primitives must be inlined before usage if needed.

// Philippe "Philo" Hurbain - 2007 - www.philohome.com

// Contains excerpts of triangle intersection code from Tomas Mller
// See thesis here: http://www.cs.lth.se/home/Tomas_Akenine_Moller/pubs/tritri.pdf
// Complete code here: http://jgt.akpeters.com/papers/Moller97/tritri.html


// Contains softSurfer Line - Triangle intersection (adaptation)
// Copyright 2001, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

#include "stdafx.h"

/* Triangle/triangle intersection test routine,
 * by Tomas Moller, 1997.
 * See article "A Fast Triangle-Triangle Intersection Test",
 * Journal of Graphics Tools, 2(2), 1997
 * updated: 2001-06-20 (added line of intersection)
 *
 * This version computes the line of intersection as well (if they are not coplanar):
 * int tri_tri_intersect_with_isectline(double V0[3],double V1[3],double V2[3], 
 *				        double U0[3],double U1[3],double U2[3],int *coplanar,
 *				        double isectpt1[3],double isectpt2[3]);
 * coplanar returns whether the tris are coplanar
 * isectpt1, isectpt2 are the endpoints of the line of intersection
 */

#include <math.h>
#include <string.h>


#define USE_EPSILON_TEST TRUE  
#define EPSILON 0.000001


/* Null vector */

double null[3] = {0,0,0};

/* some macros */

#define CROSS(dest,v1,v2)                      \
              dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
              dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
              dest[2]=v1[0]*v2[1]-v1[1]*v2[0];

#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])

#define SUB(dest,v1,v2) dest[0]=v1[0]-v2[0]; dest[1]=v1[1]-v2[1]; dest[2]=v1[2]-v2[2]; 

#define ADD(dest,v1,v2) dest[0]=v1[0]+v2[0]; dest[1]=v1[1]+v2[1]; dest[2]=v1[2]+v2[2]; 

#define MULT(dest,v,factor) dest[0]=factor*v[0]; dest[1]=factor*v[1]; dest[2]=factor*v[2];

#define SET(dest,src) dest[0]=src[0]; dest[1]=src[1]; dest[2]=src[2]; 

#define MANHATTAN(v1, v2) (fabs(v1[0]-v2[0]) + fabs(v1[1]-v2[1]) + fabs(v1[2]-v2[2]))

#define DIST(v1, v2) (sqrt((v1[0]-v2[0])*(v1[0]-v2[0]) + (v1[1]-v2[1])*(v1[1]-v2[1]) + (v1[2]-v2[2])*(v1[2]-v2[2])))


// Tri_Angle computes the cosine of the angle between the planes of two triangles.
// They are assumed to be non-degenerated
double Tri_Angle(double U0[3],double U1[3],double U2[3], double V0[3],double V1[3],double V2[3])
{
	double Unorm[3], Vnorm[3];
	double Temp[3];
	double U10[3], U20[3];
	double V10[3], V20[3];
	double len;
	SUB(U10, U1, U0);
	SUB(U20, U2, U0);
	SUB(V10, V1, V0);
	SUB(V20, V2, V0);
	CROSS(Temp, U10, U20);
	len = DIST(Temp, null);
	MULT(Unorm, Temp, 1/len);
	CROSS(Temp, V10, V20);
	len = DIST(Temp, null);
	MULT(Vnorm, Temp, 1/len);
	return(180/3.14159*acos(DOT(Unorm, Vnorm)));
}

// Check if a quad is convex and don't have aligned vertexes.
int Convex(double U0[3],double U1[3],double U2[3],double U3[3])
{
	double U01[3], U02[3], U03[3];
	double U10[3], U12[3], U13[3];
	double Temp1[3], Temp2[3];
	double c;
	SUB(U01, U1, U0);
	SUB(U02, U2, U0);
	SUB(U03, U3, U0);
	CROSS(Temp1, U01, U02);
	CROSS(Temp2, U02, U03);
	if((DOT(Temp1, Temp1)<EPSILON) ||(DOT(Temp2, Temp2)<EPSILON)) return false;
	c=DOT(Temp1, Temp2);
	if(c<0) return false;
	SUB(U10, U0, U1);
	SUB(U12, U2, U1);
	SUB(U13, U1, U3);
	CROSS(Temp1, U10, U12);
	CROSS(Temp2, U12, U13);
	if((DOT(Temp1, Temp1)<EPSILON) ||(DOT(Temp2, Temp2)<EPSILON)) return false;

	c=DOT(Temp1, Temp2);
	if(c<0) return false;

	return true;
}

#define MAX_TRI 1000
#define MAX_LINE 1000
#define SMALL 0.01
#define SMALLANGLE 0.005

void usage(void)
{
	
	printf ("Usage: ytruder [-d or -s or -p <Yval>] [-x or -y or -z] \n");
	printf ("               [-a <angle>] LdrawLineFile LdrawSurfaceFileOut\n");
	printf ("Options\n");
	printf ("  -x : Operates in x direction\n");
	printf ("  -y : Operates in y direction (default)\n");
	printf ("  -z : Operates in z direction\n");
	printf ("  -d <val>: extrude lines by a distance <val>\n");
	printf ("  -s <val>: extrude lines symmetrically vs. plane <val>.\n");
	printf ("  -p <val>: extrude down to plane <val> (projection).\n");
	printf ("  -r <val>: extrude radially by a distance <val> from direction axis.\n");
	printf ("             Default for -d/s/p <val> is -s 0 (symmetry vs. origin plane).\n");
	printf ("  -a: Conditional to fixed line conversion angle threshold <angle>.\n"); 
	printf ("      Default value is 30 degrees.\n");
	printf("Press <Enter> to quit");
	getchar();
}

int main(int argc, char* argv[])
{

	double InLine[MAX_LINE][2][3];
	int LineUsed[MAX_LINE];
	double Surf[MAX_TRI][4][3];
	double CondLine[MAX_TRI][4][3];
	int CondFlag[MAX_TRI];

	FILE *LdrawLineFile;
	FILE *LdrawSurfaceFileOut;

	int NumSurf, NumLine1, TriWritten, QuadWritten, CondWritten, LineWritten;
	int NumCond;

	char buf[1024];
	int type, color;
	int x=0, y=1, z=2;
	char dir='y';

	int mode=2;
	double AngleLineThr=30;

	int i, j, k;
	int end, current;
	int surfstart;
	
	int flag=0; 
	int direction = false;
	int sort=false;
		
	double distance = 0;

	int line = 1;

	printf("\nLdraw Extruder v1.4 - by Philo");
	printf("\n------------------------------\n\n");

	if((argc < 3)) 		
	{
		usage();
		return 1;
	}

	for(i=1; i<argc-2; i++)
	{
		if(argv[i][0] != '-') 
		{
			usage();
			return 1;
		}
		switch (argv[i][1])
		{
		case 'x':
			x=1; y=0; z=2;
			dir='x';
			break;
		case 'y':
			x=0; y=1; z=2;
			dir='y';
			break;
		case 'z':
			x=0; y=2; z=1;
			dir='z';
			break;

		case 'd':
			mode=11;
			break;
		case 's':
			mode=12;
			break;
		case 'p':
			mode=13;
			break;
		case 'r':
			mode=14;
			break;
		case 'a':
			i++;
			sscanf(argv[i], "%lf", &AngleLineThr);
			printf("Line Threshold Angle = %.1lf\n", AngleLineThr);
			break;
		default:
			usage();
			return 1;
		}
		if(mode>10)
		{
			mode=mode-10;
			i++;
			sscanf(argv[i], "%lf", &distance);
		}
	}

	switch(mode)
	{
	case 1:
		printf("Distance mode, d = %.1lf, %c direction\n", distance, dir);
		if(distance == 0)
		{
			printf("Distance = 0: nothing to do!\n");
			printf("Press <Enter> to quit");
			getchar();
			return 2;
		}
		break;
	case 2:
		printf("Symmetry mode vs. plane %c = %.1lf\n", dir, distance);
		break;
	case 3:
		printf("Projection mode on plane %c = %.1lf\n", dir, distance);
		break;
	case 4:
		printf("Radial distance mode, d = %.1lf, %c axis\n", distance, dir);
		if(distance == 0)
		{
			printf("Distance = 0: nothing to do!\n");
			printf("Press <Enter> to quit");
			getchar();
			return 2;
		}
		break;
	}

	LdrawLineFile = fopen (argv[argc-2], "rt");
	if(LdrawLineFile == NULL)
	{
		printf("Can't open file\n%s\n", argv[argc-2]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}


	LdrawSurfaceFileOut = fopen (argv[argc-1], "wt");
	if(LdrawSurfaceFileOut == NULL)
	{
		printf("Can't create file\n%s\n", argv[argc-1]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}
	
	printf("\nRead Line File...\n");
	NumLine1=0;
	while (fgets(buf, 1024, LdrawLineFile))
	{
		type = -1;
		sscanf(buf, "%d", &type);
		if(type==2)
		{
			sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf", &type, &color, 
				&InLine[NumLine1][0][x], &InLine[NumLine1][0][y], &InLine[NumLine1][0][z], 
				&InLine[NumLine1][1][x], &InLine[NumLine1][1][y], &InLine[NumLine1][1][z]);

			LineUsed[NumLine1]=0;
			NumLine1++;
		}
		if(NumLine1>MAX_LINE)
		{
			printf("Too many lines in file 1\n");
			printf("Press <Enter> to quit");
			getchar();
			return 3;
		}
	}


	printf("%d Lines in Line file\n", NumLine1);

	printf("\nExtruding\n");

	

	NumSurf=0;
	NumCond=0;
	CondFlag[NumCond]=0;
	for(i=0; i<NumLine1; i++)
	{
	double p0[3], p1[3], d0, d1;
		if(LineUsed[i] == 0)
		{
			LineUsed[i]=1;
			current=i;
			end=0;
			do
			{
				flag=false;
				for(j=0; j<NumLine1; j++)
				{
					if(LineUsed[j] == 0) 
					{
						for(k=0; k<2; k++)
						{
							if(MANHATTAN(InLine[current][end],InLine[j][k])<SMALL)
							{
								current=j;
								end=1-k;
								LineUsed[current]=1;
								flag=true;
								break;
							}
						if(flag) break;
						}
					}
				if(flag) break;
				}
			}
			while(flag);
			
			end=1-end;
			surfstart=NumSurf;
			SET(Surf[NumSurf][0], InLine[current][1-end]);
			SET(Surf[NumSurf][1], InLine[current][end]);
			SET(Surf[NumSurf][2], InLine[current][end]);
			SET(Surf[NumSurf][3], InLine[current][1-end]);
			switch (mode)
			{
			case 1 :
				Surf[NumSurf][2][1]=Surf[NumSurf][2][1] + distance;
				Surf[NumSurf][3][1]=Surf[NumSurf][3][1] + distance;
				break;
			case 2:
				Surf[NumSurf][2][1]= 2*distance - Surf[NumSurf][2][1];
				Surf[NumSurf][3][1]= 2*distance - Surf[NumSurf][3][1] ;
				break;
			case 3:
				Surf[NumSurf][2][1]=distance;
				Surf[NumSurf][3][1]=distance;
				break;
			case 4:
				p0[0]=0; p0[1]=Surf[NumSurf][0][1]; p0[2]=0; 
				p1[0]=0; p1[1]=Surf[NumSurf][1][1]; p1[2]=0; 
				d0=DIST(p0, Surf[NumSurf][0]);
				d1=DIST(p1, Surf[NumSurf][1]);
				if(d0 > EPSILON)
				{
					Surf[NumSurf][3][0] = Surf[NumSurf][3][0] * (d0+distance)/d0;
					Surf[NumSurf][3][2] = Surf[NumSurf][3][2] * (d0+distance)/d0;
				}
				if(d1 > EPSILON)
				{
					Surf[NumSurf][2][0] = Surf[NumSurf][2][0] * (d1+distance)/d1;
					Surf[NumSurf][2][2] = Surf[NumSurf][2][2] * (d1+distance)/d1;
				}
				double a;
				a=Tri_Angle(Surf[NumSurf][0], Surf[NumSurf][1], Surf[NumSurf][2], Surf[NumSurf][0], Surf[NumSurf][2], Surf[NumSurf][3]);
//				printf("%lf\n", a);
				if(a > 0.5)
				{
					SET(CondLine[NumCond][0],Surf[NumSurf][0]); 
					SET(CondLine[NumCond][1],Surf[NumSurf][2]); 
					SET(CondLine[NumCond][2],Surf[NumSurf][1]); 
					SET(CondLine[NumCond][3],Surf[NumSurf][3]); 
					CondFlag[NumCond]=5;
					NumCond++;
				}
				break;

			}
			NumSurf++;
			LineUsed[current] = 2;

			do
			{
				flag=false;
				for(j=0; j<NumLine1; j++)
				{
					if(LineUsed[j]<2) 
					{
						for(k=0; k<2; k++)
						{
							if((MANHATTAN(InLine[current][end], InLine[j][k]) <SMALL) && (LineUsed[j]<2))
							{
								current=j;
								end=1-k;
								flag=true;
								SET(Surf[NumSurf][0], InLine[current][1-end]);
								SET(Surf[NumSurf][1], InLine[current][end]);
								SET(Surf[NumSurf][2], InLine[current][end]);
								SET(Surf[NumSurf][3], InLine[current][1-end]);
								switch (mode)
								{
								case 1 :
									Surf[NumSurf][2][1]=Surf[NumSurf][2][1] + distance;
									Surf[NumSurf][3][1]=Surf[NumSurf][3][1] + distance;
									break;
								case 2:
									Surf[NumSurf][2][1]= 2*distance - Surf[NumSurf][2][1];
									Surf[NumSurf][3][1]= 2*distance - Surf[NumSurf][3][1] ;
									break;
								case 3:
									Surf[NumSurf][2][1]=distance;
									Surf[NumSurf][3][1]=distance;
									break;
								case 4:
									p0[0]=0; p0[1]=Surf[NumSurf][0][1]; p0[2]=0; 
									p1[0]=0; p1[1]=Surf[NumSurf][1][1]; p1[2]=0; 
									d0=DIST(p0, Surf[NumSurf][0]);
									d1=DIST(p1, Surf[NumSurf][1]);
									if(d0 > EPSILON)
									{
										Surf[NumSurf][3][0] = Surf[NumSurf][3][0] * (d0+distance)/d0;
										Surf[NumSurf][3][2] = Surf[NumSurf][3][2] * (d0+distance)/d0;
									}
									if(d1 > EPSILON)
									{
										Surf[NumSurf][2][0] = Surf[NumSurf][2][0] * (d1+distance)/d1;
										Surf[NumSurf][2][2] = Surf[NumSurf][2][2] * (d1+distance)/d1;
									}
									double a;
									a=Tri_Angle(Surf[NumSurf][0], Surf[NumSurf][1], Surf[NumSurf][2], Surf[NumSurf][0], Surf[NumSurf][2], Surf[NumSurf][3]);
//									printf("%lf\n", a);
									{
										SET(CondLine[NumCond][0],Surf[NumSurf][0]); 
										SET(CondLine[NumCond][1],Surf[NumSurf][2]); 
										SET(CondLine[NumCond][2],Surf[NumSurf][1]); 
										SET(CondLine[NumCond][3],Surf[NumSurf][3]); 
										CondFlag[NumCond]=5;
										NumCond++;
									}
									break;
								}
								SET(CondLine[NumCond][0],Surf[NumSurf][0]); 
								SET(CondLine[NumCond][1],Surf[NumSurf][3]); 
								SET(CondLine[NumCond][2],Surf[NumSurf][1]); 
								SET(CondLine[NumCond][3],Surf[NumSurf-1][0]); 
								CondFlag[NumCond]=5;
								NumSurf++;
								NumCond++;
								LineUsed[current] = 2;
							}
						if(flag) break;
						}
					}
					if(flag) break;
				}
			}
			while(flag);
			if(MANHATTAN(Surf[NumSurf-1][1], Surf[surfstart][0])<SMALL)
			{
				SET(CondLine[NumCond][0],Surf[NumSurf-1][1]); 
				SET(CondLine[NumCond][1],Surf[NumSurf-1][2]); 
				SET(CondLine[NumCond][2],Surf[NumSurf-1][0]); 
				SET(CondLine[NumCond][3],Surf[surfstart][1]); 
				CondFlag[NumCond]=5;
				NumCond++;
			}
			else
			{
				SET(CondLine[NumCond][0],Surf[NumSurf-1][1]); 
				SET(CondLine[NumCond][1],Surf[NumSurf-1][2]);
				CondFlag[NumCond]=2;
				NumCond++;
				SET(CondLine[NumCond][0],Surf[surfstart][0]); 
				SET(CondLine[NumCond][1],Surf[surfstart][3]); 
				CondFlag[NumCond]=2;
				NumCond++;
			}
		}
	}

	printf("%d Surfaces created\n", NumSurf);
	printf("%d Lines created\n", NumCond);

	
	printf("\nWriting output file\n");

	fprintf(LdrawSurfaceFileOut, "0 Surface %s extruded from %s\n", argv[argc-1], argv[argc-2]);
	fprintf(LdrawSurfaceFileOut, "0 Name: %s\n", argv[argc-1]);
	fprintf(LdrawSurfaceFileOut, "0 Author: Philo's Ytruder\n0\n");
//	fprintf(LdrawSurfaceFileOut, "0 BFC CERTIFY CCW\n\n");

	TriWritten=0;
	QuadWritten=0;
	CondWritten=0;
	LineWritten=0;

	for(k=0; k<NumSurf; k++)
	{
		if((MANHATTAN(Surf[k][0], Surf[k][3])<SMALL) && MANHATTAN(Surf[k][1], Surf[k][2])<SMALL) continue;
		if(MANHATTAN(Surf[k][0], Surf[k][3])<SMALL)
		{
			fprintf(LdrawSurfaceFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", Surf[k][0][x],Surf[k][0][y],Surf[k][0][z],Surf[k][1][x],Surf[k][1][y],Surf[k][1][z],Surf[k][2][x],Surf[k][2][y],Surf[k][2][z]);
			TriWritten++;
		}
		else if(MANHATTAN(Surf[k][1], Surf[k][2])<SMALL)
		{
			fprintf(LdrawSurfaceFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", Surf[k][0][x],Surf[k][0][y],Surf[k][0][z],Surf[k][1][x],Surf[k][1][y],Surf[k][1][z],Surf[k][3][x],Surf[k][3][y],Surf[k][3][z]);
			TriWritten++;
		}
		else if(Tri_Angle(Surf[k][0], Surf[k][1], Surf[k][2], Surf[k][0], Surf[k][2], Surf[k][3]) <= 0.5)

		{
			fprintf(LdrawSurfaceFileOut, "4 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", Surf[k][0][x],Surf[k][0][y],Surf[k][0][z],Surf[k][1][x],Surf[k][1][y],Surf[k][1][z],Surf[k][2][x],Surf[k][2][y],Surf[k][2][z], Surf[k][3][x],Surf[k][3][y],Surf[k][3][z]);
			QuadWritten++;
		}
		else
		{
			fprintf(LdrawSurfaceFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", Surf[k][0][x],Surf[k][0][y],Surf[k][0][z],Surf[k][1][x],Surf[k][1][y],Surf[k][1][z],Surf[k][2][x],Surf[k][2][y],Surf[k][2][z]);
			TriWritten++;
			fprintf(LdrawSurfaceFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", Surf[k][0][x],Surf[k][0][y],Surf[k][0][z],Surf[k][2][x],Surf[k][2][y],Surf[k][2][z],Surf[k][3][x],Surf[k][3][y],Surf[k][3][z]);
			TriWritten++;
		}
	}

	for(k=0; k<NumCond; k++)
	{
		if(MANHATTAN(CondLine[k][0], CondLine[k][1]) <SMALL) continue;
		if(CondFlag[k] == 5)
		{
			double a;
			a = Tri_Angle (CondLine[k][0], CondLine[k][1], CondLine[k][2], CondLine[k][0], CondLine[k][3], CondLine[k][1]);
//			printf("\n%lf\n",a); 
			if(a<AngleLineThr)
			{
				fprintf(LdrawSurfaceFileOut, "5 24 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", CondLine[k][0][x],CondLine[k][0][y],CondLine[k][0][z],CondLine[k][1][x],CondLine[k][1][y],CondLine[k][1][z],CondLine[k][2][x],CondLine[k][2][y],CondLine[k][2][z],CondLine[k][3][x],CondLine[k][3][y],CondLine[k][3][z]);
				CondWritten++;
			}
			else
			{
				fprintf(LdrawSurfaceFileOut, "2 24 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", CondLine[k][0][x],CondLine[k][0][y],CondLine[k][0][z],CondLine[k][1][x],CondLine[k][1][y],CondLine[k][1][z]);
				LineWritten++;
			}
		}

		if(CondFlag[k] == 2)
		{
			fprintf(LdrawSurfaceFileOut, "2 24 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", CondLine[k][0][x],CondLine[k][0][y],CondLine[k][0][z],CondLine[k][1][x],CondLine[k][1][y],CondLine[k][1][z]);
			LineWritten++;
		}

	}
 
	printf("%d Triangle(s) written\n", TriWritten);
	printf("%d Quad(s) written\n", QuadWritten);
	printf("%d Conditional Line(s) written\n", CondWritten);
	printf("%d Line(s) written\n", LineWritten);

	fprintf(LdrawSurfaceFileOut, "0\n");
	fclose (LdrawLineFile);
	fclose (LdrawSurfaceFileOut);
	printf("Press <Enter> to quit");
	getchar();
	return(0);
}
